@@ -29,6 +29,7 @@ import org.apache.axis2.client.Options;
 import org.apache.axis2.client.async.AxisCallback;
 import org.apache.axis2.context.MessageContext;
 import org.apache.axis2.engine.MessageReceiver;
+import org.apache.axis2.transport.http.HTTPConstants;
 import org.apache.axis2.wsdl.WSDLConstants;
 import org.apache.commons.logging.Log;
 import org.apache.commons.logging.LogFactory;
@@ -107,6 +108,26 @@ public class SynapseCallbackReceiver implements MessageReceiver {
 
         String messageID = null;
 
+        /**
+         * In an Out-only scenario if the client receives a HTTP 202 accepted we need to
+         * remove the call back/s registered for that request.
+         * This if will check weather this is a message sent in a that scenario and remove the callback
+         */
+        if (messageCtx.getProperty(NhttpConstants.HTTP_202_RECEIVED) != null && "true".equals(
+                messageCtx.getProperty(NhttpConstants.HTTP_202_RECEIVED))) {
+            if (callbackStore.containsKey(messageCtx.getMessageID())) {
+                callbackStore.remove(messageCtx.getMessageID());
+                if (log.isDebugEnabled()) {
+                    log.debug("CallBack registered with Message id : " + messageCtx.getMessageID() +
+                            " removed from the " +
+                            "callback store since we got an accepted Notification");
+                }
+            }
+
+            return;
+        }
+
+
         if (messageCtx.getOptions() != null && messageCtx.getOptions().getRelatesTo() != null) {
             // never take a chance with a NPE at this stage.. so check at each level :-)
             Options options = messageCtx.getOptions();
@@ -180,7 +201,7 @@ public class SynapseCallbackReceiver implements MessageReceiver {
                     synapseOutMsgCtx.getEnvelope().build();
                 } catch (OMException x) {
                     synapseOutMsgCtx.setEnvelope(response.getEnvelope());
-                  }
+                }
 
                 Exception e = (Exception) response.getProperty(SynapseConstants.ERROR_EXCEPTION);
 
@@ -207,7 +228,7 @@ public class SynapseCallbackReceiver implements MessageReceiver {
             // there can always be only one instance of an Endpoint in the faultStack of a message
             // if the send was successful, so remove it before we proceed any further
             Stack faultStack = synapseOutMsgCtx.getFaultStack();
-            if (faultStack !=null && !faultStack.isEmpty()
+            if (faultStack != null && !faultStack.isEmpty()
                 && faultStack.peek() instanceof Endpoint) {
                 Endpoint successfulEndpoint = (Endpoint) faultStack.pop();
                 successfulEndpoint.onSuccess();
@@ -284,10 +305,15 @@ public class SynapseCallbackReceiver implements MessageReceiver {
                 response.removeProperty(AddressingConstants.DISABLE_ADDRESSING_FOR_OUT_MESSAGES);
             }
 
-             // copy the message type property thats used by the out message to the response message
-            response.setProperty(org.apache.axis2.Constants.Configuration.MESSAGE_TYPE,
-                axisOutMsgCtx.getProperty(org.apache.axis2.Constants.Configuration.MESSAGE_TYPE));
-            
+            Object messageType = axisOutMsgCtx.getProperty(
+                    org.apache.axis2.Constants.Configuration.MESSAGE_TYPE);
+            if (!HTTPConstants.MEDIA_TYPE_X_WWW_FORM.equals(messageType)) {
+                 // copy the message type property that's used by the out message to the
+                 // response message
+                response.setProperty(org.apache.axis2.Constants.Configuration.MESSAGE_TYPE,
+                    messageType);
+            }
+
             // compare original received message (axisOutMsgCtx) soap version with the response
             // if they are different change to original version 
             if(axisOutMsgCtx.isSOAP11() != response.isSOAP11()) {
