@@ -18,11 +18,10 @@
  */
 package org.apache.synapse.transport.nhttp;
 
+import org.apache.axiom.om.OMAbstractFactory;
 import org.apache.axiom.soap.SOAP11Constants;
 import org.apache.axiom.soap.SOAP12Constants;
 import org.apache.axiom.soap.SOAPEnvelope;
-import org.apache.axiom.soap.impl.llom.soap11.SOAP11Factory;
-import org.apache.axiom.soap.impl.llom.soap12.SOAP12Factory;
 import org.apache.axis2.AxisFault;
 import org.apache.axis2.Constants;
 import org.apache.axis2.addressing.AddressingConstants;
@@ -61,8 +60,12 @@ import org.apache.synapse.transport.nhttp.util.NhttpMetricsCollector;
 import org.apache.synapse.commons.jmx.ThreadingView;
 
 import java.io.IOException;
+import java.net.URI;
+import java.net.URISyntaxException;
+import java.util.Comparator;
 import java.util.Map;
 import java.util.HashMap;
+import java.util.TreeMap;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.locks.Lock;
 import java.util.concurrent.locks.ReentrantLock;
@@ -153,6 +156,9 @@ public class ClientHandler implements NHttpClientHandler {
         this.countConnections = NHttpConfiguration.getInstance().isCountConnections();
         // set the connection map to the configuration context
         cfgCtx.setProperty(NhttpConstants.OPEN_CONNNECTIONS_MAP, openConnections);
+
+        // set the latest openConnections map to MBean data during connection creation
+        metrics.setConnectionsPerHosts(openConnections);
     }
 
     public void requestReady(final NHttpClientConnection conn) {
@@ -192,7 +198,7 @@ public class ClientHandler implements NHttpClientHandler {
         if (countConnections) {
             recordConnection(conn);
         }
-        
+
         try {
             processConnection(conn, (Axis2HttpRequest) attachment);
         } catch (ConnectionClosedException e) {
@@ -542,11 +548,13 @@ public class ClientHandler implements NHttpClientHandler {
                 // reset metrics on connection
                 conn.getMetrics().reset();
                 if (context.getAttribute(NhttpConstants.DISCARD_ON_COMPLETE) != null) {
-                    // this is a connection we should not re-use
-                    ConnectionPool.forget(conn);
-                    shutdownConnection(conn);
-                    context.removeAttribute(RESPONSE_SINK_BUFFER);
-                    context.removeAttribute(REQUEST_SOURCE_BUFFER);                    
+                    try {
+                        // this is a connection we should not re-use
+                        ConnectionPool.forget(conn);
+                        shutdownConnection(conn);
+                        context.removeAttribute(RESPONSE_SINK_BUFFER);
+                        context.removeAttribute(REQUEST_SOURCE_BUFFER);
+                    } catch (Exception ignore) {}
                 } else if (!connStrategy.keepAlive(response, context)) {
                     shutdownConnection(conn);
                     context.removeAttribute(RESPONSE_SINK_BUFFER);
@@ -630,6 +638,9 @@ public class ClientHandler implements NHttpClientHandler {
      */
     public void responseReceived(final NHttpClientConnection conn) {
 
+        setServerContextAttribute(NhttpConstants.RES_HEADER_ARRIVAL_TIME,
+                System.currentTimeMillis(), conn);
+
         HttpContext context = conn.getContext();
         HttpResponse response = conn.getHttpResponse();
 
@@ -711,11 +722,26 @@ public class ClientHandler implements NHttpClientHandler {
                         MessageContext responseMsgCtx = outMsgCtx.getOperationContext().
                                 getMessageContext(WSDL2Constants.MESSAGE_LABEL_IN);
                         if (responseMsgCtx == null ||
+                                outMsgCtx.getOptions().isUseSeparateListener() ||
+                                outMsgCtx.getOperationContext().isComplete()) {
+                            if (responseMsgCtx != null &&
+                                    responseMsgCtx.getProperty("synapse.send") == null) {
+                                return;
+                            }
+                        } else if (responseMsgCtx == null ||
                                 outMsgCtx.getOptions().isUseSeparateListener()) {
-                            // This means that we received a 202 accepted for an out-only ,
-                            // for which we do not need a dummy message anyway
+                            // Since we need to notify the SynapseCallback receiver to remove the
+                            // call backs registered  we set a custom property
+                            setHeaders(context, response, outMsgCtx, responseMsgCtx);
+                            outMsgCtx.setProperty(NhttpConstants.HTTP_202_RECEIVED, "true");
+                            mr.receive(outMsgCtx);
                             return;
                         }
+
+                        if (responseMsgCtx == null) {
+                            return;
+                        }
+                        setHeaders(context, response, outMsgCtx, responseMsgCtx);
                         responseMsgCtx.setServerSide(true);
                         responseMsgCtx.setDoingREST(outMsgCtx.isDoingREST());
                         responseMsgCtx.setProperty(MessageContext.TRANSPORT_IN,
@@ -730,9 +756,9 @@ public class ClientHandler implements NHttpClientHandler {
                         responseMsgCtx.setTo(null);
 
                         if (!outMsgCtx.isDoingREST() && !outMsgCtx.isSOAP11()) {
-                            responseMsgCtx.setEnvelope(new SOAP12Factory().getDefaultEnvelope());
+                            responseMsgCtx.setEnvelope(OMAbstractFactory.getSOAP12Factory().getDefaultEnvelope());
                         } else {
-                            responseMsgCtx.setEnvelope(new SOAP11Factory().getDefaultEnvelope());
+                            responseMsgCtx.setEnvelope(OMAbstractFactory.getSOAP11Factory().getDefaultEnvelope());
                         }
                         responseMsgCtx.setProperty(AddressingConstants.
                                 DISABLE_ADDRESSING_FOR_OUT_MESSAGES, Boolean.TRUE);
@@ -802,6 +828,53 @@ public class ClientHandler implements NHttpClientHandler {
         }
     }
 
+    private void setHeaders(HttpContext context, HttpResponse response,
+                            MessageContext outMsgCtx, MessageContext responseMsgCtx) {
+        Header[] headers = response.getAllHeaders();
+        if (headers != null && headers.length > 0) {
+
+            Map<String, String> headerMap
+                    = new TreeMap<String, String>(new Comparator<String>() {
+                public int compare(String o1, String o2) {
+                    return o1.compareToIgnoreCase(o2);
+                }
+            });
+            String endpointURLPrefix = (String) context.getAttribute(NhttpConstants.ENDPOINT_PREFIX);
+
+            String servicePrefix = (String) outMsgCtx.getProperty(NhttpConstants.SERVICE_PREFIX);
+            for (int i = 0; i < headers.length; i++) {
+                Header header = headers[i];
+                if ("Location".equals(header.getName())
+                        && endpointURLPrefix != null && servicePrefix != null) {
+                    //Here, we are changing only the host name and the port of the new URI - value of the Location
+                    //header.
+                    //If the new URI is again referring to a resource in the server to which the original request
+                    //is sent, then replace the hostname and port of the URI with the hostname and port of synapse
+                    //We are not changing the request url here, only the host name and the port.
+                    try {
+                        URI serviceURI = new URI(servicePrefix);
+                        URI endpointURI = new URI(endpointURLPrefix);
+                        URI locationURI = new URI(header.getValue());
+
+                        if (locationURI.getHost().equalsIgnoreCase(endpointURI.getHost())) {
+                            URI newURI = new URI(locationURI.getScheme(), locationURI.getUserInfo(),
+                                    serviceURI.getHost(), serviceURI.getPort(), locationURI.getPath(),
+                                    locationURI.getQuery(), locationURI.getFragment());
+                            headerMap.put(header.getName(), newURI.toString());
+                            responseMsgCtx.setProperty(NhttpConstants.SERVICE_PREFIX,
+                                    outMsgCtx.getProperty(NhttpConstants.SERVICE_PREFIX));
+                        }
+                    } catch (URISyntaxException e) {
+                        log.error(e.getMessage(), e);
+                    }
+                } else {
+                    headerMap.put(header.getName(), header.getValue());
+                }
+            }
+            responseMsgCtx.setProperty(MessageContext.TRANSPORT_HEADERS, headerMap);
+        }
+    }
+
     /**
      * Checks whether the provided 500 response shall be logged as a warning.
      * The behavior can be configured based on the content type of the message via a transport 
